// # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
// ang2nc.c
// by Jian Zheng (jian.zheng@nasa.gov)
//
// Description:
//   ang2nc was developed and tested on Linux Redhat 8.5 (gcc 8.5).
//   This program ingests Aviris-NG datasets and format the data into
//   netCDF-4/HDF5 files. It will process all the valid input files in the working
//   directory.
//
//   If *_rdn, *_rdn.hdr, *_igm, *_igm.hdr, *_obs are present,
//      a AVIRISngL1B_<FlightDate>_<start HHMM>_<end HHMM>.nc file will be generated.
//
//   If *_rdn, *_rdn.hdr are present, either *_igm and/or *_obs file are missing
//      a *_l1.nc file will be generated.
//
//   If *_rfl, *_rfl.hdr, *_igm, *_igm.hdr, *_obs  are present,
//      a AVIRISngL2_<FlightDate>_<start HHMM>_<end HHMM>.nc file will be generated.
//
//   If *_rfl, *_rfl.hdr are present, either *_igm and/or *_obs file are missing
//      a *_l2.nc file will be generated.
//
// Compilation:
//   To compile the program on Linux system, use the following command,
//     gcc -o ang2nc ang2nc.c -lnetcdf -lm
//   This program requires netcdf library to be installed. The program
//   was developed using netcdf.x86_64 v4.7 found on the EPEL Linux repository.
//
// Syntax:
//    ang2nc <Path>
//    The <Path> name is optional if the program is called in the current working
//    directory.
// # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

/*
G.Wind Changes made to the code

Slope is now an int16 SDS with a scale factor of 0.01
FlightStartTime and FlightEndTime attributes have been added
EarthSunDistance has been removed as it is easily computed on the fly from date/time

*/
/* Dec 7th, 2023
 * ang2nc will process OBS and IGM data in bil or bip format. It can also handle OBS data
 * in ENVI data type 4 (32-bit float) or 5 (64-bit double).
 */

#include <stdio.h>    // always needed
#include <unistd.h>   // For PC Linux
#include <dirent.h>   // always needed for directory listing
#include <string.h>   // needed for strcpy, strcat, ...
#include <stdlib.h>   // needed for 'system', _MAX_*, ...
#include <time.h>     // needed for clock
#include <malloc.h>   // for calloc
#include <sys/stat.h> // for file size greatere than 2G
#include <math.h>
#include <netcdf.h>

#define MAX_INT  65536
#define Samples  598
#define Bands    425
#define OBSBands 11
#define DPI      3.14159265358979323846

/* For the units attributes. */
#define UNITS   "units"
#define DEF     "Description"
#define NANO    "nanometer"
#define MT      "meter"
#define DEG     "degree"
#define HOUR    "hour"
#define AU      "astronomical unit"
#define RDNDATA "uW nm-1 cm-2 sr-1"
#define SCALE   "scale_factor"
#define OFFSET  "add_offset"

// Handle errors by printing an error message and exiting with a
// non-zero status. 
#define ERRCODE 2
#define ERR(e) {printf("Error: %s\n", nc_strerror(e)); exit(ERRCODE);}

#define MIN(a,b) (((a)<(b))?(a):(b))
// # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
// # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
//              L O C A L   F U N C T I O N   P R O T O T Y P E S             //
// # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
// # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

typedef   struct  {
  char CreationDate[32];
  char Title[128];
  char FlightDate[16];
  char FlightStartTime[8];
  char FlightEndTime[8];
  char RadianceVersion[8];
  char BadPixelMap[128];
  char FlatFieldFile[128];
  char RCCFile[128];
  char SpectralScatterFile[128];
  char WavelengthFile[128];
  char CrossTrackScatterFile[128];
  int  UTMzone;
  int  Granule_Size;
} GlobalAttr;

char *GA[] =  {"CreationDate",           //0
               "Title",                  //1
               "FlightDate",             //2
               "RadianceVersion",        //3
               "BadPixelMap",            //4
               "FlatFieldFile",          //5
               "RCCFile",                //6
               "SpectralScatterFile",    //7
               "WavelengthFile",         //8
               "CrossTrackScatterFile",  //9
               "UTMzone", 				 //10
			   "FlightStartTime",        //11
			   "FlightEndTime"};			 //12
               //"Granule_Size"};          //11

char *SDS[] = {"Wavelength",             //0
               "FWHM",                   //1
			   "PathLength",             //2
			   "SensorAzimuth",          //3
			   "SensorZenith",           //4
			   "SunAzimuth",             //5
			   "SunZenith",              //6
			   "SolarPhase",             //7
			   "Slope",                  //8
			   "Aspect",                 //9
			   "Cosine(i)",              //10
			   "CoordinatedUniversalTime(UTC)",  //11
			   "EarthSunDistance",       //12
			   "PixelLatitude",          //13
			   "PixelLongitude",         //14
			   "PixelEasting",           //15
			   "PixelNorthing",          //16
			   "PixelElevation",         //17
               "CalibratedData"} ;       //18
			   
char *SDS2[] = {"Wavelength",             //0
                "FWHM",                   //1
			    "PixelLatitude",          //2
			    "PixelLongitude",         //3
                "CoordinatedUniversalTime(UTC)",  //4
                "ReflectanceData"};       //5

//===========================================
void mygetline(FILE*, char*);
int  getLines(FILE *);
void readIGMhdr(FILE *, GlobalAttr *, int *, int *);
void readOBShdr(FILE *, int *, int *, int *);
void readRxxhdr(FILE *, GlobalAttr *, float *, float *); // works for rdn or rfl hdr file
void readUTC(FILE *, FILE *, int, char *, char *);
void writeGA(int, GlobalAttr *, int, int);
int  rwIGM(FILE *, int, int, int, int, int, int *, int, double *, double *);
void writeLatLon(int, int, int *, double *, double *);
int  rwRxx(FILE *, int, int, int *, int);
int  rwOBS(FILE *, FILE *, int, int, int *, int);
void tm_to_latlon(int zone, int hemi, double* lat, double* lon, double* x, double* y, int pix);
off_t getGranuleSize(char *filename);
int minLines(FILE *rdnhdr, FILE *igmhdr, FILE* obshdr);
// # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
// # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


int main(int argc, char** argv)
{
    char path[64];
    char root[64], root2[64];
    char rdnName[128], igmName[128], obsName[128], rflName[128];
    char rdnHdrName[128], igmHdrName[128], obsHdrName[128], rflHdrName[128];

    // For directory listing
    char *nx, *token;
    struct dirent **fptr;
	int i, n, index;
		
    FILE *igmfile = NULL, *igmhdr = NULL;
    FILE *obsfile = NULL, *obshdr = NULL;
    FILE *rdnfile = NULL, *rdnhdr = NULL;
    FILE *rflfile = NULL, *rflhdr = NULL;

    time_t currentTime;
    int status;
    int lines, igm_lines, obs_lines, diff=0;  // lines in this flight line
    int igm_interleave;
    int hemi;   // North = 1, South = 0
    size_t ix, jx;
    GlobalAttr glbattr, glbattr2;
	char L1bfilename[64];
	char L2filename[64];
	char YYYYMMDD[16];
	char startHHMM[8];
	char endHHMM[8];
	
	// netCDF variables, groups, demensions, etc
	int l1bfid, l1b_grp_sds;    // L1B file and group ids
	int l2fid, l2_grp_sds;       // L2 file and group ids
	int band_dim, sample_dim, line_dim;
    int dim2id[2], dim3id[3];	
	int SDS_varid[18], SDS2_varid[6];

	float scale_factor[1];
	float add_offset[1];


	char *tsis_name;
	float *tsis_data=NULL;
	int tsis_id, idum;
	FILE* f;
	char buffer[200], dum[50];
	float fdum;


    char temp[128];
    long long granuleSize;

    //data sets
    float *wavelen = calloc(Bands, sizeof(float) );
    float *fwhm = calloc(Bands, sizeof(float) );
    double *lat = NULL, *lon = NULL;
    int latlon = 0, igm = 0, obs = 0;


    printf("# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #\n");
    printf("# Syntax: ang2nc <Path>                                                           #\n");
    printf("#                                                                                 #\n");
    printf("# This program will format Aviris-NG data into netCDF-4/HDF5 files.               #\n");
    printf("#                                                                                 #\n");
    printf("# If *_rdn, *_rdn.hdr, *_igm, *_igm.hdr, *_obs are present,                       #\n");
    printf("#    a AVIRISngL1B_<FlightDate>_<start HHMM>_<end HHMM>.nc file will be generated.#\n");
    printf("#                                                                                 #\n");
    printf("# If *_rdn, *_rdn.hdr are present, either *_igm and/or *_obs file are missing     #\n");
    printf("#    a *_l1.nc file will be generated.                                            #\n");
    printf("#                                                                                 #\n");
    printf("# If *_rfl, *_rfl.hdr, *_igm, *_igm.hdr, *_obs  are present,                      #\n");
    printf("#    a AVIRISngL2_<FlightDate>_<start HHMM>_<end HHMM>.nc file will be generated. #\n");
    printf("#                                                                                 #\n");
    printf("# If *_rfl, *_rfl.hdr are present, either *_igm and/or *_obs file are missing     #\n");
    printf("#    a *_l2.nc file will be generated.                                            #\n");
    printf("# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #\n");

    // List _rdn_v2aa1_clip.hdr to get file root name.
    // There may or may not be a Level-2 reflectance file.
    // Process the files in a while loop

    if(argc < 3) sprintf(path, "."); // for tsis_name
//    if(argc < 2) sprintf(path, ".");
    else strcpy( path, argv[1] );

    index = scandir(path, &fptr, NULL, alphasort);
	if(n < 0) {
		perror("scandir");
		exit(EXIT_FAILURE);
	}

	tsis_name = argv[2];

// not so fast, rdn can also be a part of IGM and OBS filenames. 
	for(n = 0; n < index; n++){
		if ( (nx=strstr(fptr[n]->d_name, "rdn_v")) != NULL ) { // Found radiance header file
            // Make certain that this is rdn data file, not the hdr file
            strcpy(temp, fptr[n]->d_name);
            if( (token = strstr(temp, ".hdr")) == NULL) { // Not the _rdn_*.hdr file
                // In case there are added char behind _rdn
                sprintf(rdnName,"%s/%s", path, temp);
                sprintf(rdnHdrName,"%s.hdr", rdnName);
                printf("\nProcessing %s...\n", rdnName);
            }
            else { // It's the _rdn_*.hdr file
                continue;
            }

            token = strtok(fptr[n]->d_name, "_");
            strcpy(root, token);

            // Open all 5 files needed for the L1B. If any file is missing, skip.

            if( ( rdnhdr = fopen( rdnHdrName, "r" ) ) == NULL ) {
                printf( "%s fopen failure\n", rdnHdrName );
                goto L1DONE;
            }

            if( ( rdnfile = fopen( rdnName, "r" ) ) == NULL ) {
                printf( "%s fopen failure\n", rdnName );
                printf( "Skipping %s L1b output\n", root);
                goto L1DONE;
            }


// G.Wind 8.8.23 The logic for selecting and opening IGM and OBS files has been redone, because
// apparently those files can come in different filename conventions. And manually renaming files is just ewww

            igm = 0;
            sprintf(igmHdrName, "%s/%s%s", path,root, "_igm.hdr");
			igmhdr = fopen( igmHdrName, "r" ); 
			if (igmhdr == NULL) {
	            sprintf(igmHdrName, "%s/%s%s", path,root, "_rdn_igm.hdr");
				igmhdr = fopen( igmHdrName, "r" );
			}
			if (igmhdr == NULL) {
	            sprintf(igmHdrName, "%s/%s%s", path,root, "_ort_igm.hdr");
				igmhdr = fopen( igmHdrName, "r" );
			}

            sprintf(igmName, "%s/%s%s", path,root, "_igm");
			igmfile = fopen( igmName, "r" );
			if (igmfile == NULL) {
	            sprintf(igmName, "%s/%s%s", path,root, "_rdn_igm");
				igmfile = fopen( igmName, "r" );
			}
            if (igmfile == NULL) {
	            sprintf(igmName, "%s/%s%s", path,root, "_ort_igm");
				igmfile = fopen( igmName, "r" );
			}

            if( igmhdr == NULL || igmfile == NULL ) {
                printf( "IGM fopen failure\n" );
                printf( "%s L1b output will not have IGM data\n", root);
            }
            else igm = 1;



            obs = 0;
            sprintf(obsHdrName, "%s/%s%s", path,root, "_obs.hdr");
			obshdr = fopen( obsHdrName, "r" );
			if (obshdr == NULL) {
	            sprintf(obsHdrName, "%s/%s%s", path,root, "_rdn_obs.hdr");
				obshdr = fopen( obsHdrName, "r" );
			}
            sprintf(obsName, "%s/%s%s", path, root, "_obs");
			obsfile = fopen( obsName, "r" );
			if (obsfile == NULL) {
	            sprintf(obsName, "%s/%s%s", path, root, "_rdn_obs");
				obsfile = fopen( obsName, "r" );
			}

            if ( obsfile == NULL ) {
                printf( "%s fopen failure\n", obsName );
                printf( "%s L1b output will not have OBS data\n", root);
                //goto L1DONE;
            }
            else obs = 1;

			
            // Read Hdr for line number, find the smallest, then allocate for data arrays
            lines = minLines(rdnhdr, igmhdr, obshdr);

            // Fill in GlobalAttr
            strcpy( glbattr.Title, "Airborne Visible Infrared Imaging Spectrometer Next Generation (AVIRIS NG) Level-1B Data");
            strncpy(glbattr.FlightDate, &root[3], 8);
			glbattr.FlightDate[8]='\0';

			if(obs == 1 && igm == 1){
			    // Read *_obs for first and last UTC time
			    readUTC(obsfile, obshdr, lines, startHHMM, endHHMM);

				sprintf(glbattr.FlightStartTime, "%s", startHHMM);
				sprintf(glbattr.FlightEndTime, "%s", endHHMM);

			    // Get YYYYMMDD from input file name angYYYYMMDDt..._
			    sprintf(L1bfilename, "%s/AVIRISngL1B_%s_%s_%s.nc", path,glbattr.FlightDate, startHHMM, endHHMM);
                printf("L1b output file name: %s \n", L1bfilename);
            }
            else {
                sprintf(L1bfilename, "%s/%s_l1.nc", path, root);
                printf("L1b output file name: %s \n", L1bfilename);
            }
//printf("obsName = %s\n", obsName);
//printf("igmName = %s\n", igmName);
            //CreationDate
            time( &currentTime );  // get time as long integer
            ix = sizeof( glbattr.CreationDate );  // ix should be 32, which is more than enough
            jx = strftime( glbattr.CreationDate, ix, "%d-%b-%Y %H:%M:%S",
                             localtime( &currentTime ) );  // jx should equal 20

			// Create the netCDF 4 file
			if((status = nc_create(L1bfilename, NC_NETCDF4|NC_CLOBBER, &l1bfid)))
				ERR(status);
			
			// Create Scientific Data Sets group
			status = nc_def_grp(l1bfid, "Scientific_Data_Sets", &l1b_grp_sds);
			if(status != NC_NOERR)	ERR(status);

            // Get global attributes
            if(igm == 1)
                readIGMhdr(igmhdr, &glbattr, &hemi, &igm_interleave);

			readRxxhdr(rdnhdr, &glbattr, wavelen, fwhm);

            writeGA(l1bfid, &glbattr, igm, 1);

	        // Define SDS dimensions and variables
            status = nc_def_dim(l1b_grp_sds, "Band_dim", Bands, &band_dim);
            if (status != NC_NOERR) ERR(status);
            // 1-D
			for(ix = 0; ix < 2; ix++){
				status = nc_def_var(l1b_grp_sds, SDS[ix], NC_FLOAT, 1, &band_dim, &SDS_varid[ix]);
                if(status != NC_NOERR) ERR(status);
			}

            // 2-D
            status = nc_def_dim(l1b_grp_sds, "Sample_dim", Samples, &sample_dim);
            if (status != NC_NOERR) ERR(status);
            status = nc_def_dim(l1b_grp_sds, "Lines_dim", lines, &line_dim);
            if (status != NC_NOERR) ERR(status);
            dim2id[0] = line_dim;
            dim2id[1] = sample_dim;

            if(obs == 1){ //If OBS exists make the following SDS, otherwise don't
			    for(ix = 2; ix < 12; ix++){ // 2-D float
					if (ix==3 || ix==4 || ix == 5 || ix==6 || ix == 8 || ix == 9) 
		                status = nc_def_var(l1b_grp_sds, SDS[ix], NC_SHORT, 2, dim2id, &SDS_varid[ix]); // Slope and aspect should be INT2
					else
		                status = nc_def_var(l1b_grp_sds, SDS[ix], NC_FLOAT, 2, dim2id, &SDS_varid[ix]);


                    if(status != NC_NOERR) ERR(status);
                }

//                status = nc_def_var(l1b_grp_sds, SDS[12], NC_DOUBLE, 2, dim2id, &SDS_varid[12]);
//                if(status != NC_NOERR) ERR(status);
            }

            if(igm == 1){ //If IMG exists make the following SDS, otherwise don't
                for(ix = 13; ix < 18; ix++){ // 2-D double
                    status = nc_def_var(l1b_grp_sds, SDS[ix], NC_DOUBLE, 2, dim2id, &SDS_varid[ix]);
                    if(status != NC_NOERR) ERR(status);
                }
            }

            // 3-D
            dim3id[0] = line_dim; //line_dim3;
			dim3id[1] = band_dim; //sample_dim3;
			dim3id[2] = sample_dim;

            status = nc_def_var(l1b_grp_sds, SDS[18], NC_FLOAT, 3, dim3id, &SDS_varid[18]);
            if(status != NC_NOERR) ERR(status);


//JZ: don't have this file for testing
// Add the solar spectral irradiance from TSIS to AVIRIS files
            status = nc_def_var(l1b_grp_sds, "Solar_Spectral_Irradiance", NC_FLOAT, 1, &band_dim, &tsis_id);
            if(status != NC_NOERR) ERR(status);
			sprintf(temp, "%s", "uW / cm^2 / nm");
            status = nc_put_att_text(l1b_grp_sds, tsis_id, "units", strlen(temp), temp);

            tsis_data = (float *)calloc(Bands, sizeof(float) );

			
			f = fopen(tsis_name, "r");
			for (i=0; i<10; i++)
				mygetline(f, buffer);
	
			for (i=0; i<Bands; i++) {
				mygetline(f, buffer);
				sscanf(buffer, "%s %d %d %f %f %f", dum, &idum, &idum, &fdum, &fdum, &(tsis_data[i]));
				tsis_data[i] = tsis_data[i] / 10.0;
			}

			fclose(f);

            status = nc_put_var_float(l1b_grp_sds, tsis_id, tsis_data);


			free(tsis_data);
// end TSIS additions




	        // Define variable attributes
            //Wavelength
            status = nc_put_att_text(l1b_grp_sds, SDS_varid[0], UNITS, strlen(NANO), NANO);
            if(status != NC_NOERR) ERR(status);
            //fwhm
            status = nc_put_att_text(l1b_grp_sds, SDS_varid[1], UNITS, strlen(NANO), NANO);
            if(status != NC_NOERR) ERR(status);

            if(obs == 1){
                //PathLength
                status = nc_put_att_text(l1b_grp_sds, SDS_varid[2], UNITS, strlen(MT), MT);
                if(status != NC_NOERR) ERR(status);
                //SensorAzimuth
                status = nc_put_att_text(l1b_grp_sds, SDS_varid[3], UNITS, strlen(DEG), DEG);
                if(status != NC_NOERR) ERR(status);
				scale_factor[0] = 0.01;
                status = nc_put_att_float(l1b_grp_sds, SDS_varid[3], SCALE, NC_FLOAT, 1, scale_factor);
				add_offset[0] = 180.0;
                status = nc_put_att_float(l1b_grp_sds, SDS_varid[3], OFFSET, NC_FLOAT, 1, add_offset);
                sprintf(temp,"%s", "0 to 360 degrees CW from N");
                status = nc_put_att_text(l1b_grp_sds, SDS_varid[3], DEF, strlen(temp), temp);
                if(status != NC_NOERR) ERR(status);
                //SensorZenith
                status = nc_put_att_text(l1b_grp_sds, SDS_varid[4], UNITS, strlen(DEG), DEG);
                if(status != NC_NOERR) ERR(status);
				scale_factor[0] = 0.01;
                status = nc_put_att_float(l1b_grp_sds, SDS_varid[4], SCALE, NC_FLOAT, 1, scale_factor);
                sprintf(temp,"%s", "0 to 90 degrees from Zenith");
                status = nc_put_att_text(l1b_grp_sds, SDS_varid[4], DEF, strlen(temp), temp);
                if(status != NC_NOERR) ERR(status);
                //SunAzimuth
                status = nc_put_att_text(l1b_grp_sds, SDS_varid[5], UNITS, strlen(DEG), DEG);
                if(status != NC_NOERR) ERR(status);
				scale_factor[0] = 0.01;
                status = nc_put_att_float(l1b_grp_sds, SDS_varid[5], SCALE, NC_FLOAT, 1, scale_factor);
				add_offset[0] = 180.0;
                status = nc_put_att_float(l1b_grp_sds, SDS_varid[5], OFFSET, NC_FLOAT, 1, add_offset);
                sprintf(temp,"%s", "0 to 360 degrees CW from N");
                status = nc_put_att_text(l1b_grp_sds, SDS_varid[5], DEF, strlen(temp), temp);
                if(status != NC_NOERR) ERR(status);
                //SunZenith
                status = nc_put_att_text(l1b_grp_sds, SDS_varid[6], UNITS, strlen(DEG), DEG);
                if(status != NC_NOERR) ERR(status);
				scale_factor[0] = 0.01;
                status = nc_put_att_float(l1b_grp_sds, SDS_varid[6], SCALE, NC_FLOAT, 1, scale_factor);
                sprintf(temp,"%s", "0 to 90 degrees from Zenith");
                status = nc_put_att_text(l1b_grp_sds, SDS_varid[6], DEF, strlen(temp), temp);
                if(status != NC_NOERR) ERR(status);
                //SolarPhase
                status = nc_put_att_text(l1b_grp_sds, SDS_varid[7], UNITS, strlen(DEG), DEG);
                if(status != NC_NOERR) ERR(status);
                sprintf(temp,"%s", "The degrees between to-sensor and to-sun vectors in the principal plane");
                status = nc_put_att_text(l1b_grp_sds, SDS_varid[7], DEF, strlen(temp), temp);
                if(status != NC_NOERR) ERR(status);
                //Slope
                status = nc_put_att_text(l1b_grp_sds, SDS_varid[8], UNITS, strlen(DEG), DEG);
				scale_factor[0] = 0.01;
                status = nc_put_att_float(l1b_grp_sds, SDS_varid[8], SCALE, NC_FLOAT, 1, scale_factor);
                if(status != NC_NOERR) ERR(status);
                sprintf(temp,"%s", "Local surface slope as derived from the DEM");
                status = nc_put_att_text(l1b_grp_sds, SDS_varid[8], DEF, strlen(temp), temp);
                if(status != NC_NOERR) ERR(status);
                //Aspect
                status = nc_put_att_text(l1b_grp_sds, SDS_varid[9], UNITS, strlen(DEG), DEG);
                if(status != NC_NOERR) ERR(status);
				scale_factor[0] = 0.01;
                status = nc_put_att_float(l1b_grp_sds, SDS_varid[9], SCALE, NC_FLOAT, 1, scale_factor);
				add_offset[0] = 180.0;
                status = nc_put_att_float(l1b_grp_sds, SDS_varid[9], OFFSET, NC_FLOAT, 1, add_offset);
                sprintf(temp,"%s", "Local surface aspect 0 to 360 degrees cw from N");
                status = nc_put_att_text(l1b_grp_sds, SDS_varid[9], DEF, strlen(temp), temp);
                if(status != NC_NOERR) ERR(status);
                //Cosine
                sprintf(temp,"%s", "Apparent local illumination factor based on DEM slope and aspect to-sun vector, -1 to 1");
                status = nc_put_att_text(l1b_grp_sds, SDS_varid[10], DEF, strlen(temp), temp);
                if(status != NC_NOERR) ERR(status);
                //UTC
                status = nc_put_att_text(l1b_grp_sds, SDS_varid[11], UNITS, strlen(HOUR), HOUR);
                if(status != NC_NOERR) ERR(status);
                //Earth-sun
//                status = nc_put_att_text(l1b_grp_sds, SDS_varid[12], UNITS, strlen(AU), AU);
//                if(status != NC_NOERR) ERR(status);
            }

            if(igm == 1){
			    status = nc_put_att_text(l1b_grp_sds, SDS_varid[13], UNITS, strlen(DEG), DEG);
                if(status != NC_NOERR) ERR(status);

                status = nc_put_att_text(l1b_grp_sds, SDS_varid[14], UNITS, strlen(DEG), DEG);
                if(status != NC_NOERR) ERR(status);

			    status = nc_put_att_text(l1b_grp_sds, SDS_varid[15], UNITS, strlen(MT), MT);
                if(status != NC_NOERR) ERR(status);

			    status = nc_put_att_text(l1b_grp_sds, SDS_varid[16], UNITS, strlen(MT), MT);
                if(status != NC_NOERR) ERR(status);

			    status = nc_put_att_text(l1b_grp_sds, SDS_varid[17], UNITS, strlen(MT), MT);
                if(status != NC_NOERR) ERR(status);
            }

            //RDN
			status = nc_put_att_text(l1b_grp_sds, SDS_varid[18], UNITS, strlen(RDNDATA), RDNDATA);
            if(status != NC_NOERR) ERR(status);

			// End define mode
			status = nc_enddef(l1bfid);
            if(status != NC_NOERR) ERR(status);

	        // Read/Write SDS data
            status = nc_put_var_float(l1b_grp_sds, SDS_varid[0], wavelen);
            if(status != NC_NOERR) ERR(status);

            status = nc_put_var_float(l1b_grp_sds, SDS_varid[1], fwhm);
            if(status != NC_NOERR) ERR(status);

            if(obs == 1){
	            status = rwOBS(obsfile, obshdr, lines, l1b_grp_sds, SDS_varid, 1);
                if(status < 0) printf("Error in R/W OBS data...");
            }

            //Save lat/lon for RFL L2 output
            if(igm == 1){
                latlon = 0;
                lat = (double *)calloc(Samples*lines, sizeof(double) );
                lon = (double *)calloc(Samples*lines, sizeof(double) );
                if(lat != NULL && lon != NULL) {
	                status = rwIGM(igmfile, lines, igm_interleave, glbattr.UTMzone, hemi, l1b_grp_sds, SDS_varid, 1, lat, lon);
                    if(status < 0) printf("Error in R/W IGM data...");
                    else latlon = 1;
                }
                else latlon = 0;
            }

	        status = rwRxx(rdnfile, lines, l1b_grp_sds, SDS_varid, 1);
            if(status < 0) printf("Error in R/W IGM data...");

            // Close nc file
            if((status = nc_close(l1bfid)))
                ERR(status);

            granuleSize = (long long)getGranuleSize(L1bfilename);

            if(granuleSize > 0) {
                if((status = nc_open(L1bfilename, NC_NETCDF4|NC_WRITE, &l1bfid)))
				    ERR(status);

                status = nc_put_att_longlong(l1bfid, NC_GLOBAL, "Granule_Size", NC_INT64, 1, &granuleSize);
                    if(status != NC_NOERR) ERR(status);

                // End define mode
			    status = nc_enddef(l1bfid);
                if(status != NC_NOERR) ERR(status);

                // Close nc file
			    if((status = nc_close(l1bfid)))
                    ERR(status);
            }

L1DONE:
            printf("Close %s L1b input files...\n", root);
            if(rdnhdr  != NULL) {fclose(rdnhdr);  rdnhdr  = NULL;}
            if(rdnfile != NULL) {fclose(rdnfile); rdnfile = NULL;}
            if(igmhdr  != NULL) {fclose(igmhdr);  igmhdr  = NULL;}
            if(igmfile != NULL) {fclose(igmfile); igmfile = NULL;}
            if(obsfile != NULL) {fclose(obsfile); obsfile = NULL;}
            if(obshdr  != NULL) {fclose(obshdr);  obshdr  = NULL;}

		} //End if(... "rdn" ... )
        else if ( (nx=strstr(fptr[n]->d_name, "rfl")) != NULL ) { // Found radiance header file
            // Make certain that this is rfl data file, not the hdr file
            strcpy(temp, fptr[n]->d_name);
            if( (token = strstr(temp, ".hdr")) == NULL) {
                sprintf(rflName, "%s/%s", path, temp);
                sprintf(rflHdrName,"%s.hdr", rflName);
                printf("\nProcessing %s...\n", rflName);
            }
            else { // It's the rfl_*.hdr file
                continue; // Since the fptr is alphasort'ed, _rfl come up first
            }

            token = strtok(fptr[n]->d_name, "_");
            strcpy(root2, token);

            // Check to see if there are reflectance files for Level 2A HDF
            if( ( rflhdr = fopen( rflHdrName, "r" ) ) == NULL ) {
                printf( "%s fopen failure\n", rflHdrName );
                printf( "Skipping %s L2 processing\n", rflHdrName);
                goto L2DONE;
            }

            if( ( rflfile = fopen( rflName, "r" ) ) == NULL ) {
                printf( "%s fopen failure\n", rflName );
                printf( "Skipping %s L2 processing\n", rflName);
                goto L2DONE;
            }

            obs = 0;
            sprintf(obsHdrName, "%s/%s%s", path,root2, "_obs.hdr");
            if( ( obshdr = fopen( obsHdrName, "r" ) ) == NULL ) {
                printf( "%s fopen failure\n", obsHdrName );
                printf( "%s L2 output will not have OBS data\n", root2);
                //goto L2DONE;
            }
            else {
                sprintf(obsName, "%s/%s%s", path, root2, "_obs");
                if( ( obsfile = fopen( obsName, "r" ) ) == NULL ) {
                    printf( "%s fopen failure\n", obsName );
                    printf( "%s L2 output will not have OBS data\n", root2);
                    //goto L2DONE;
                }
                else obs = 1;
            }

            igm = 0;
            if(latlon == 0) { // If latlon exist from L1b process
                sprintf(igmHdrName, "%s/%s%s", path, root2, "_igm.hdr");
                if( ( igmhdr = fopen( igmHdrName, "r" ) ) == NULL ) {
                    printf( "%s fopen failure\n", igmHdrName );
                    printf( "%s L2 output will not have IGM data\n", root2);
                    //goto L2DONE;
                }
                else { // If hdr don't exist, no need to open IGM
                    sprintf(igmName, "%s/%s%s", path, root2, "_igm");
                    if( ( igmfile = fopen( igmName, "r" ) ) == NULL ) {
                        printf( "%s fopen failure\n", igmName );
                        printf( "%s L2 output will not have IGM data\n", root2);
                        //goto L2DONE;
                    }
                    else igm = 1;
                }
            }
            //else if(latlon == 1)
            //    printf("Skipping IMG: using Lat/Lon data generated while processing RND file\n");

			// Read Hdr for line number, find the smallest, then allocate for data arrays
            lines = minLines(rflhdr, igmhdr, obshdr);


            strncpy(glbattr2.FlightDate, &root2[3], 8);
            glbattr2.FlightDate[8]='\0';

			if(obs == 1) {
			    // Read *_obs for first and last UTC time
			    readUTC(obsfile, obshdr, lines, startHHMM, endHHMM);
            }

            if(obs == 1 && (latlon == 1 || igm == 1) ) {
			    sprintf(L2filename, "%s/AVIRISngL2_%s_%s_%s.nc", path, glbattr2.FlightDate, startHHMM, endHHMM);
			    printf("L2 output file name: %s \n", L2filename);
            }
            else{
                sprintf(L2filename, "%s/%s_l2.nc", path,root2);
			    printf("L2 output file name: %s \n", L2filename);
            }

            //CreationDate
            time( &currentTime );  // get time as long integer
            ix = sizeof( glbattr2.CreationDate );
            jx = strftime( glbattr2.CreationDate, ix, "%d-%b-%Y %H:%M:%S",
                             localtime( &currentTime ) );  // jx should equal 20
							 
			// Fill in GlobalAttr
            strcpy(glbattr2.Title, "Airborne Visible Infrared Imaging Sepctrometer Next Generation (AVIRIS NG) Level-2 Data");

			if((status = nc_create(L2filename, NC_NETCDF4|NC_CLOBBER, &l2fid)))
				ERR(status);

			// Create Scientific Data Sets group
			status = nc_def_grp(l2fid, "Scientific_Data_Sets", &l2_grp_sds);
			if(status != NC_NOERR)	ERR(status);

            // Define global attributes
            if(latlon == 0 && igm == 1)
                readIGMhdr(igmhdr, &glbattr2, &hemi, &igm_interleave);

			readRxxhdr(rflhdr, &glbattr2, wavelen, fwhm);

            writeGA(l2fid, &glbattr2, igm, 2);

			// Define SDS dimensions and variables
            status = nc_def_dim(l2_grp_sds, "Band_dim", Bands, &band_dim);
            if (status != NC_NOERR) ERR(status);
            // 1-D
			for(ix = 0; ix < 2; ix++){
				status = nc_def_var(l2_grp_sds, SDS2[ix], NC_FLOAT, 1, &band_dim, &SDS2_varid[ix]);
                if(status != NC_NOERR) ERR(status);
			}

            // 2-D
            status = nc_def_dim(l2_grp_sds, "Sample_dim", Samples, &sample_dim);
            if (status != NC_NOERR) ERR(status);
            status = nc_def_dim(l2_grp_sds, "Lines_dim", lines, &line_dim);
            if (status != NC_NOERR) ERR(status);
	        dim2id[0] = line_dim;
            dim2id[1] = sample_dim;

            if(latlon == 1 || igm == 1){
			    for(ix = 2; ix < 4; ix++){ // lat, lon
	                status = nc_def_var(l2_grp_sds, SDS2[ix], NC_DOUBLE, 2, dim2id, &SDS2_varid[ix]);
                    if(status != NC_NOERR) ERR(status);
                }
            }
            //UTC
            if(obs == 1) {
                status = nc_def_var(l2_grp_sds, SDS2[4], NC_FLOAT, 2, dim2id, &SDS2_varid[4]);
                if(status != NC_NOERR) ERR(status);
            }
            // 3-D
            dim3id[0] = line_dim;
			dim3id[1] = band_dim;
			dim3id[2] = sample_dim;
			status = nc_def_var(l2_grp_sds, SDS2[5], NC_FLOAT, 3, dim3id, &SDS2_varid[5]);
            if(status != NC_NOERR) ERR(status);

			// Define variable attributes
            //Wavelength
			status = nc_put_att_text(l2_grp_sds, SDS2_varid[0], UNITS, strlen(NANO), NANO);
            if(status != NC_NOERR) ERR(status);
            //fwhm
            status = nc_put_att_text(l2_grp_sds, SDS2_varid[1], UNITS, strlen(NANO), NANO);
            if(status != NC_NOERR) ERR(status);
            if(latlon == 1 || igm == 1){
                // lat
			    status = nc_put_att_text(l2_grp_sds, SDS2_varid[2], UNITS, strlen(DEG), DEG);
                if(status != NC_NOERR) ERR(status);
			    // lon
                status = nc_put_att_text(l2_grp_sds, SDS2_varid[3], UNITS, strlen(DEG), DEG);
                if(status != NC_NOERR) ERR(status);
            }
            if(obs == 1){
                //UTC
                status = nc_put_att_text(l2_grp_sds, SDS2_varid[4], UNITS, strlen(HOUR), HOUR);
                if(status != NC_NOERR) ERR(status);
            }
			// End define mode
			status = nc_enddef(l2fid);
            if(status != NC_NOERR) ERR(status);

            // Read/Write SDS data
            status = nc_put_var_float(l2_grp_sds, SDS2_varid[0], wavelen);
            if(status != NC_NOERR) ERR(status);

            status = nc_put_var_float(l2_grp_sds, SDS2_varid[1], fwhm);
            if(status != NC_NOERR) ERR(status);

            // Use lat/lon saved from L1b
            if(latlon == 1) { // Have latlon, write to file
                //printf("Existing lat/lon lat = (%f)%p, lon = (%f)%p\n", lat[0], lat, lon[0],lon);
                writeLatLon(l2_grp_sds, lines, SDS2_varid, lat, lon);
            }
            else if(igm == 1){ // Don't have existing lat/lon from L1b, compute latlon
                status = rwIGM(igmfile, lines, igm_interleave, glbattr.UTMzone, hemi, l2_grp_sds, SDS2_varid, 2, lat, lon);
                if(status < 0) printf("Error in R/W IGM data...");
            }
            if(obs == 1){
                // Only needs UTC
                status = rwOBS(obsfile, obshdr, lines, l2_grp_sds, SDS2_varid, 2);
                if(status < 0) printf("Error in R/W OBS data...");
            }
            status = rwRxx(rflfile, lines, l2_grp_sds, SDS2_varid, 2);
            if(status < 0) printf("Error in R/W IGM data...");
						
			// Close nc file
			if((status = nc_close(l2fid)))
                ERR(status);

            granuleSize = (long long)getGranuleSize(L2filename);

            if(granuleSize > 0) {
                if((status = nc_open(L2filename, NC_NETCDF4|NC_WRITE, &l2fid)))
				    ERR(status);

                status = nc_put_att_longlong(l2fid, NC_GLOBAL, "Granule_Size", NC_INT64, 1, &granuleSize);
                    if(status != NC_NOERR) ERR(status);

                // End define mode
			    status = nc_enddef(l2fid);
                if(status != NC_NOERR) ERR(status);

                // Close nc file
			    if((status = nc_close(l2fid)))
                    ERR(status);
            }
L2DONE:
            printf("Close %s L2 input files...\n", root2);
            if(obshdr  != NULL) {fclose(obshdr);  obshdr  = NULL;}
            if(obsfile != NULL) {fclose(obsfile); obsfile = NULL;}
            if(igmfile != NULL) {fclose(igmfile); igmfile = NULL;}
            if(igmhdr  != NULL) {fclose(igmhdr);  igmhdr  = NULL;}
            if(rflfile != NULL) {fclose(rflfile); rflfile = NULL;}
            if(rflhdr  != NULL) {fclose(rflhdr);  rflhdr  = NULL;}
            if(lat != NULL) {free(lat); lat = NULL;}
            if(lon != NULL) {free(lon); lon = NULL;}

        } // end else if( )

    } // end for(n = 0; n < index; n++)

    if(wavelen != NULL) free(wavelen);
    if(fwhm != NULL)    free(fwhm);

    for(i = 0; i < index; i++) free(fptr[i]);
    free(fptr);

    return 0;
} // end main( )

void mygetline(FILE*f, char* buffer){
    char c;
    int i=0;
    while (!feof(f)){
        if ( (c=getc(f)) != '\n') {
            buffer[i] = c;
            i++;
        }
        else 
            break;
    }
    buffer[i] = '\0';
}

//===========================================
int getLines( FILE *hdrfile )
{
    char buff[64], name[16];
    char *str, *val;
    int lines;
    int status;

    status = fseek(hdrfile, 0, SEEK_SET);

    while(fgets( buff, 64, hdrfile ) != NULL){
        str = strtok(buff, "=");
        strcpy(name, str);
        if(strncmp(name, "lines", 5) == 0){
            val = strtok(NULL, "");
            lines = atoi(val);

            break;
        }
    }
    status = fseek(hdrfile, 0, SEEK_SET);

    return lines;
}

//===========================================
void readUTC(FILE *obsfile, FILE *obshdr, int lines, char *startHHMM, char *endHHMM)
{
	double startTime = 0.0, endTime = 0.0;
	float fstartTime = -999.0, fendTime = -999.0;
	int status;
	long long offset = 0;
	int startHH, startMM;
	int endHH, endMM;
    int bands, type, interleave;

    readOBShdr(obshdr, &bands, &type, &interleave);

    status = fseek(obsfile, 0, SEEK_SET);
    if(interleave == 1){ // bil
        // UTC is the 10th data set
        if(type == 5) {// 64-bit double
            offset = sizeof(double) * Samples * 9;        // Skipping 9 bands of the 1st line
            status = fseek(obsfile, offset, SEEK_SET);
            fread(&startTime, sizeof(double), 1, obsfile);
            //printf("(1/5)startTime = %f", startTime);

            // Read end time
            status = fseek(obsfile, 0, SEEK_SET);
            offset = sizeof(double) * Samples * (lines - 1) * bands;  // Skip to the last line
            offset += sizeof(double) * Samples * 9;           // Skip to the UTC band
            offset += sizeof(double) * (Samples - 1);         // Before the last sample
            status = fseek(obsfile, offset, SEEK_SET);
            if(status == 0) fread(&endTime, sizeof(double), 1, obsfile);
            //printf("(1/5)endTime = %f\n", endTime);
        }
        else if(type == 4){ //32-bit float
            offset = sizeof(float) * Samples * 9;        // Skipping 9 bands of the 1st line
            status = fseek(obsfile, offset, SEEK_SET);
            fread(&fstartTime, sizeof(float), 1, obsfile);
            //printf("(1/4)startTime = %f ", fstartTime);

            // Read end time
            status = fseek(obsfile, 0, SEEK_SET);
            offset = sizeof(float) * Samples * (lines - 1) * bands;  // Skip to the last line
            offset += sizeof(float) * Samples * 9;           // Skip to the UTC band
            offset += sizeof(float) * (Samples - 1);         // Before the last sample
            status = fseek(obsfile, offset, SEEK_SET);
            if(status == 0) fread(&fendTime, sizeof(float), 1, obsfile);
            //printf("(1/4)endTime = %f \n",fendTime);
        }
    }
    else if(interleave == 0) { // bip
        if(type == 5) { //64-bit double
            offset = sizeof(double) * 9;    // Skipping the 1st 9 values of the 1st pixel
            status = fseek(obsfile, offset, SEEK_SET);
            if(status == 0) fread(&startTime, sizeof(double), 1, obsfile);
            //printf("(0/5)startTime = %f ", startTime);

            // Read end time
            status = fseek(obsfile, 0, SEEK_SET);
            offset = sizeof(double) * Samples * (lines-1) * bands;    // Skipping 9 data sets
            offset += sizeof(double) * (Samples-1) *bands;       // Skip to the last pixel
            offset += sizeof(double) * 9;                           // Skip 9 values
            status = fseek(obsfile, offset, SEEK_SET);
            if(status == 0) fread(&endTime, sizeof(double), 1, obsfile);
            //printf("(0/5)endTime = %f\n", endTime);
        }
        else if(type == 4){ //32-bit float
            offset = sizeof(float) * 9;    // Skipping the 1st 9 values of the 1st pixel
            status = fseek(obsfile, offset, SEEK_SET);
            if(status == 0) fread(&fstartTime, sizeof(float), 1, obsfile);
            //printf("(0/4)startTime = %f ", fstartTime);

            // Read end time
            status = fseek(obsfile, 0, SEEK_SET);
            offset = sizeof(float) * Samples * (lines-1) * bands;    // Skipping 9 data sets
            offset += sizeof(float) * (Samples-1) *bands;       // Skip to the last pixel
            offset += sizeof(float) * 9;                           // Skip 9 values
            status = fseek(obsfile, offset, SEEK_SET);
            if(status == 0) fread(&fendTime, sizeof(float), 1, obsfile);
            //printf("(0/4)endTime = %f\n", fendTime);
        }
    }
     if(fstartTime > 0) {
        startTime = (double)fstartTime;
        endTime = (double)fendTime;
     }
    //printf("startTime = %f, endTime = %f\n");
    startHH = (int)startTime;
    startMM = (int)floor((startTime - startHH) * 60);
    //sprintf(startHHMM, "%d%d", startHH, startMM);
    if(startHH < 10) sprintf(startHHMM, "0%d", startHH);
    else sprintf(startHHMM, "%d", startHH);
    if(startMM < 10) sprintf(startHHMM, "%s0%d", startHHMM, startMM);
    else sprintf(startHHMM, "%s%d", startHHMM, startMM);

    endHH = (int)endTime;
    endMM = (int)floor((endTime - endHH) * 60);
    //sprintf(endHHMM, "%d%d", endHH, endMM);
    if(startHH < 10) sprintf(endHHMM, "0%d", endHH);
    else sprintf(endHHMM, "%d", endHH);
    if(endMM < 10) sprintf(endHHMM, "%s0%d", endHHMM,endMM);
    else sprintf(endHHMM, "%s%d", endHHMM, endMM);

	status = fseek(obsfile, 0, SEEK_SET);
	return;
}

//===========================================
void readOBShdr(FILE *obshdr, int *bands, int *type, int *interleave)
{
    char *buff =  calloc(9600, sizeof(char));
    char buff2[256], name[32];
    char *str, *val;
    int status;

    while(fgets( buff, 9600, obshdr ) != NULL){
        str = strtok(buff, "=");
        strcpy(name, str);

        if(strncmp(name, "bands", 5) == 0){
			val = strtok(NULL, "");
            *bands = atoi(val);
        }
        else if(strncmp(name, "data type", 9) == 0){
            val = strtok(NULL, "");
            *type = atoi(val);
        }
        else if(strncmp(name, "interleave", 10) == 0){
            val = strtok(NULL, "");
            if(strstr(val, "bip") != NULL)  *interleave = 0;
            else if(strstr(val, "bil") != NULL)  *interleave = 1;
        }
    }

    status = fseek(obshdr, 0, SEEK_SET); //reading it more than once
    free(buff);
    return;
}

//===========================================
void readRxxhdr(FILE *rxxhdr, GlobalAttr *ga, float *wavelen, float *fwhm)
{
    int ix, jx, bands;
    char *buff =  calloc(9600, sizeof(char)); // some lines are really long
    char buff2[256], name[32];
    char *str, *val;

    while(fgets( buff, 9600, rxxhdr ) != NULL){
        str = strtok(buff, "=");
        strcpy(name, str);
		
		if(strncmp(name, "bands", 5) == 0){
			val = strtok(NULL, "");
            bands = atoi(val);
        }
        else if(strncmp(name, "radiance version", 16) == 0){
            val = strtok(NULL, "");
            strcpy(buff2, val);
            for(ix = 0; ix < 5; ix++){
                if(buff2[ix] != ' '){ // find the first char of the string
                    strcpy(ga->RadianceVersion, &(buff2[ix]));
                    //printf("%s\n", ga->RadianceVersion);
                    break;
                }
            }
        }
        else if(strncmp(name, "bad pixel map", 13) == 0){
            val = strtok(NULL, "");
            strcpy(buff2, val);
            for(ix = 0; ix < 5; ix++){
                if(buff2[ix] != ' '){
                    strcpy(ga->BadPixelMap, &(buff2[ix]));
                    break;
                }
            }
        }
        else if(strncmp(name, "flat field file", 15) == 0){
            val = strtok(NULL, "");
            strcpy(buff2, val);
            for(ix = 0; ix < 5; ix++){
                if(buff2[ix] != ' '){
                    strcpy(ga->FlatFieldFile, &(buff2[ix]));
                    break;
                }
            }
        }
        else if(strncmp(name, "rcc file", 8) == 0){
            val = strtok(NULL, "");
            strcpy(buff2, val);
            for(ix = 0; ix < 5; ix++){
                if(buff2[ix] != ' '){
                    strcpy(ga->RCCFile, &(buff2[ix]));
                    break;
                }
            }
        }
        else if(strncmp(name, "spectral scatter file", 21) == 0){
            val = strtok(NULL, "");
            strcpy(buff2, val);
            for(ix = 0; ix < 5; ix++){
                if(buff2[ix] != ' '){
                    strcpy(ga->SpectralScatterFile, &(buff2[ix]));
                    break;
                }
            }
        }
        else if(strncmp(name, "wavelength file", 15) == 0){
            val = strtok(NULL, "");
            strcpy(buff2, val);
            for(ix = 0; ix < 5; ix++){
                if(buff2[ix] != ' '){
                    strcpy(ga->WavelengthFile, &(buff2[ix]));
                    break;
                }
            }
        }
        else if(strncmp(name, "crosstrack scatter file", 23) == 0){
            val = strtok(NULL, "");
            strcpy(buff2, val);
            for(ix = 0; ix < 5; ix++){ // Skip white space
                if(buff2[ix] != ' '){
                    strcpy(ga->CrossTrackScatterFile, &(buff2[ix]));
                    //printf("%s\n", glbattr->CrossTrackScatterFile);
                    break;
                }
            }
        }
		else if(strncmp(name, "wavelength", 10) == 0){  //new
            val = strtok(NULL, "{");
			for(ix = 0; ix < bands; ix++){
			    if(ix == bands - 1) val = strtok(NULL, "}");
				else val = strtok(NULL, ",");

				strcpy(buff2, val);
			    wavelen[ix] = (float)atof(buff2);
				//if(ix==0 || ix == bands-1) printf("wavelength[ix] = %f\n", ix,wavelen[ix]);
			}
		}
		else if(strncmp(name, "fwhm", 4) == 0){  //new
            val = strtok(NULL, "{");
			for(ix = 0; ix < bands; ix++){
			    if(ix == bands - 1)  val = strtok(NULL, "}");
				else val = strtok(NULL, ",");

				strcpy(buff2, val);
			    fwhm[ix] = (float)atof(buff2);
				//if(ix==0 || ix == bands-1) printf("fwhm[ix] = %f\n", ix,fwhm[ix]);
			}
			break; // break from while, all inputs found
		}
			
    } // end while( )
    free(buff);
    return;
}

//===========================================
void readIGMhdr(FILE *igmhdr, GlobalAttr *glbattr, int *hemi, int *igm_interleave)
{
    char buff[256], buff2[256], name[32];
    char *str, *val;
    int status;
	
    while(fgets( buff, 256, igmhdr ) != NULL){
        strcpy(buff2, buff);

        str = strtok(buff, " ");
        strcpy(name, str); //UTM
        if(strncmp(name, "UTM", 3) == 0) {
            str = strtok(NULL, " "); // zone
            val = strtok(NULL, " "); // number
            glbattr->UTMzone = atoi(val);
            str = strtok(NULL, "}"); // North

            if( strcmp(str, "North") == 0) *hemi = 1;
            else *hemi = 0;
            //printf("UTM zone = %d, hemi = %d \n", glbattr->UTMzone, *hemi);
            continue;
        }
        // For lines with delimited by "=", such as "interleave = bip"
        str = strtok(buff2, "=");
        strcpy(name, str);

		if(strncmp(name, "interleave", 10) == 0){
			val = strtok(NULL, "");
            if(strstr(val, "bip") != NULL)  *igm_interleave = 0;
            else if(strstr(val, "bil") != NULL)  *igm_interleave = 1;
        }

    } // end while( )

    status = fseek(igmhdr, 0, SEEK_SET);
    return;
}

//===========================================
void writeGA(int lfid, GlobalAttr *glbattr, int igm, int level)
{
    int status;
    size_t len;

    len = strlen(glbattr->CreationDate);
    status = nc_put_att_text(lfid, NC_GLOBAL, GA[0], len, glbattr->CreationDate);
    if(status != NC_NOERR) ERR(status);

    len = strlen(glbattr->Title);
    status = nc_put_att_text(lfid, NC_GLOBAL, GA[1], len, glbattr->Title);
    if(status != NC_NOERR) ERR(status);

    len = strlen(glbattr->FlightDate);
    status = nc_put_att_text(lfid, NC_GLOBAL, GA[2], len, glbattr->FlightDate);
    if(status != NC_NOERR) ERR(status);

    len = strlen(glbattr->FlightStartTime);
    status = nc_put_att_text(lfid, NC_GLOBAL, GA[11], len, glbattr->FlightStartTime);
    if(status != NC_NOERR) ERR(status);
    len = strlen(glbattr->FlightEndTime);
    status = nc_put_att_text(lfid, NC_GLOBAL, GA[12], len, glbattr->FlightEndTime);
    if(status != NC_NOERR) ERR(status);



    len = strlen(glbattr->RadianceVersion);
    status = nc_put_att_text(lfid, NC_GLOBAL, GA[3], len, glbattr->RadianceVersion);
    if(status != NC_NOERR) ERR(status);

    len = strlen(glbattr->BadPixelMap);
    status = nc_put_att_text(lfid, NC_GLOBAL, GA[4], len, glbattr->BadPixelMap);
    if(status != NC_NOERR) ERR(status);

    len = strlen(glbattr->FlatFieldFile);
    status = nc_put_att_text(lfid, NC_GLOBAL, GA[5], len, glbattr->FlatFieldFile);
    if(status != NC_NOERR) ERR(status);

    len = strlen(glbattr->RCCFile);
    status = nc_put_att_text(lfid, NC_GLOBAL, GA[6], len, glbattr->RCCFile);
    if(status != NC_NOERR) ERR(status);

    len = strlen(glbattr->SpectralScatterFile);
    status = nc_put_att_text(lfid, NC_GLOBAL, GA[7], len, glbattr->SpectralScatterFile);
    if(status != NC_NOERR) ERR(status);

    len = strlen(glbattr->WavelengthFile);
    status = nc_put_att_text(lfid, NC_GLOBAL, GA[8], len, glbattr->WavelengthFile);
    if(status != NC_NOERR) ERR(status);

    len = strlen(glbattr->CrossTrackScatterFile);
    status = nc_put_att_text(lfid, NC_GLOBAL, GA[9], len, glbattr->CrossTrackScatterFile);
    if(status != NC_NOERR) ERR(status);
    
    if(igm == 1 && level == 1){
        status = nc_put_att_int(lfid, NC_GLOBAL, GA[10], NC_INT, 1, &(glbattr->UTMzone));
        if(status != NC_NOERR) ERR(status);
	}

}

//===========================================
int rwRxx(FILE *rxxfile, int lines, int grp_sds, int *varid, int level)
{
    int i, status;
	size_t start[3], count[3];
    float *buff = calloc(Bands*Samples, sizeof(float));
    int lineSize = Samples*Bands*sizeof(float);
	
	count[0] = 1;
    count[1] = Bands;
    count[2] = Samples;
    start[1] = 0;
    start[2] = 0;

    for (i = 0; i < lines; i++){
        if(  ( fread( buff, lineSize, 1, rxxfile ) ) != 1 ) {
            printf("Error reading Rxx input\n");
            return -1;
        }

        start[0] = i;
        if(level == 1){
            status = nc_put_vara_float(grp_sds, varid[18], start, count, buff);
            if(status != NC_NOERR) ERR(status);

        }
        else if(level == 2){ //Aviris L2
            status = nc_put_vara_float(grp_sds, varid[5], start, count, buff);
            if(status != NC_NOERR) ERR(status);
        }
    }

    free(buff);
    return 1;
}

//===========================================
int rwOBS(FILE *obsfile, FILE *obshdr, int lines, int l1b_grp_sds, int *SDS_varid, int level)
{
    int i, j, l,status, x, y;
    int bands, type, interleave;
    size_t start[2], count[2];
    double *buff = calloc(Samples, sizeof(double));
    float  *fbuff = calloc(Samples, sizeof(float));
	short *sbuff = calloc(Samples, sizeof(short));
    int dlineSize = Samples*sizeof(double);
    int flineSize = Samples*sizeof(float);
    double *tempDBuf = NULL;
    float *tempFBuf = NULL;

    count[0] = 1;
    count[1] = Samples;
    start[1] = 0;

    // if level = 1 (L1b), inputs are double, so covert to float
    //                     except last one (Earth-sun) as double
    // if level = 2 (L2),  inputs are double, so convert to float
    //                     only write UTC (data set 11)

    // These AVIRIS obs files can have 10 or 11 bands, 32-b or 64-b data type,
    // and bip or bil interleave
    readOBShdr(obshdr, &bands, &type, &interleave);
    if(type == 5) tempDBuf = calloc(bands*Samples, sizeof(double));
    else if(type == 4) tempFBuf = calloc(bands*Samples, sizeof(float));

    for(i = 0; i < lines; i++) {
        start[0] = i;

        // For interleave bip format, read all the bands, sort them into bil format later
        if(interleave == 0) {//bip
            if(type == 5) {
                if(  ( fread( tempDBuf, bands*dlineSize, 1, obsfile ) ) != 1 ) {
                    printf("Error reading OBS input bip/5\n");
                    return -1;
                }
            }
            else if(type == 4) {
               if(  ( fread( tempFBuf, bands*flineSize, 1, obsfile ) ) != 1 ) {
                    printf("Error reading OBS input bip/4\n");
                    return -1;
                }
            }
        }

        // For bil, read one band at a time. For bip, sort one band at a time into bil
        for(j = 2; j < 12; j++) {
            if(interleave == 1) { // bil
                if(type == 5) {
                    if(  ( fread( buff, dlineSize, 1, obsfile ) ) != 1 ) {
                        printf("Error reading OBS input bil/5\n");
                        return -1;
                    }
                }
                else if(type == 4) {
                    if(  ( fread( fbuff, flineSize, 1, obsfile ) ) != 1 ) {
                        printf("Error reading OBS input bil/4\n");
                        return -1;
                    }
                    for(y = 0; y < Samples; y++)
                        buff[y] = (double)fbuff[y];
                }
            }
            else if(interleave == 0) { // sort bip into bil format
                for(x = 0; x < Samples; x++) {
                    if(type == 5) buff[x] = tempDBuf[x*bands + (j-2)];
                    else if(type == 4) buff[x] = (double)tempFBuf[x*bands + (j-2)];

                    //if(i == 0 )//debug
                     //   printf("type(%d): buff[%d] = tempBuf[%d*%d + %d] (%d)\n",
                      //              type,       x,          x,bands,(j-2), x*bands+j-2 );
                }
            }

            if(level == 1){ // L1b uses all the data sets

				if (j==4 || j==6 || j==8 ) {
	                for(l = 0; l < Samples; l++)
	                    sbuff[l] = (short)(buff[l]/0.01);

	                // Write to netCDF file
	                status = nc_put_vara_short(l1b_grp_sds, SDS_varid[j], start, count, sbuff);

				}else if (j==3 || j==5 || j==9) {
	                for(l = 0; l < Samples; l++)
	                    sbuff[l] = (short)((buff[l]-180.)/0.01);

	                // Write to netCDF file
	                status = nc_put_vara_short(l1b_grp_sds, SDS_varid[j], start, count, sbuff);
	
				} else {

	                for(l = 0; l < Samples; l++)
	                    fbuff[l] = (float)buff[l];

	                // Write to netCDF file
	                status = nc_put_vara_float(l1b_grp_sds, SDS_varid[j], start, count, fbuff);
	
				}
	            if(status != NC_NOERR) ERR(status);
            }
            else if(level == 2 && j == 11){ // L2 only needs UTC
                for(l = 0; l < Samples; l++)
                    fbuff[l] = (float)buff[l];

                // Write UTC
                status = nc_put_vara_float(l1b_grp_sds, SDS_varid[4], start, count, fbuff);
                if(status != NC_NOERR) ERR(status);
            }
        } // end for(j = ...)

        // Reading the 12th SDS data set EarthSunDistance
        if(interleave == 1 && bands == 11){ //only for bil format
            if(type == 5) {
                if(  ( fread( buff, dlineSize, 1, obsfile ) ) != 1 ) {
                    printf("Error reading OBS input bil/5 11th band\n");
                    return -1;
                }
            }
            else if(type == 4) {
                if(  ( fread( fbuff, flineSize, 1, obsfile ) ) != 1 ) {
                    printf("Error reading OBS input bil/4 11th band\n");
                    return -1;
                }
            }
        }
//        if(level == 1){ //EarthSunDistance
//            // Write to netCDF file
//            status = nc_put_vara_double(l1b_grp_sds, SDS_varid[12], start, count, buff);
//            if(status != NC_NOERR) ERR(status);
//        }
    } // end for(i = ...)

	if(tempDBuf != NULL) free(tempDBuf);
    if(tempFBuf != NULL) free(tempFBuf);
    free(buff); free(fbuff);
	free(sbuff);
    return 1;
}

//===========================================
int rwIGM(FILE *igmfile, int lines, int interleave, int zone, int hemi, int grp_sds, int *varid, int level, double *lat, double *lon)
{
    int i, j, k, status, x;
    size_t lineSize = Samples*sizeof(double);
    double *east = calloc(Samples, sizeof(double));
    double *north= calloc(Samples, sizeof(double));
    double *elev = calloc(Samples, sizeof(double));
    double *latln = calloc(Samples, sizeof(double));
    double *lonln = calloc(Samples, sizeof(double));
    double *tempBuf = calloc(3*Samples, sizeof(double));
    long long offset;

    size_t start[2], count[2];

    count[0] = 1;
    count[1] = Samples;
    start[1] = 0;

    for(i = 0; i < lines; i++){
        if(interleave == 1) {  //bil
            if( ( fread( east, lineSize, 1, igmfile ) ) != 1 ) {
                printf("Error reading bil IGM input\n");
                return -1;
            }
            if( ( fread( north, lineSize, 1, igmfile ) ) != 1 ) {
                printf("Error reading bil IGM input\n");
                return -1;
            }
            if( ( fread( elev, lineSize, 1, igmfile ) ) != 1 ) {
                printf("Error reading bil IGM input\n");
                return -1;
            }
        }
        else if(interleave == 0){ //bip
            if( ( fread( tempBuf, 3*lineSize, 1, igmfile ) ) != 1 ) {
                printf("Error reading bip IGM input\n");
                return -1;
            }
            for(j = 0; j < Samples; j++) {
                k = j*3;
                east[j] = tempBuf[k];
                north[j] = tempBuf[k+1];
                elev[j] = tempBuf[k+2];
            }
        }
/*
if(i == 0){
  for(x =0; x < 10; x++)
    printf("east[%d] = %f, north[%d] = %f, elev[%d] = %f\n", x, east[x], x,north[x], x, elev[x]);
  for(x =290; x < 300; x++)
    printf("east[%d] = %f, north[%d] = %f, elev[%d] = %f\n", x, east[x], x,north[x], x, elev[x]);
  for(x =580; x < 590; x++)
    printf("east[%d] = %f, north[%d] = %f, elev[%d] = %f\n", x, east[x], x,north[x], x, elev[x]);
}
*/
        tm_to_latlon(zone, hemi, latln, lonln, east, north, Samples);

        // Write to netCDF file
        start[0] = i;
        
        if(level == 1) {
            status = nc_put_vara_double(grp_sds, varid[13], start, count, latln);
            if(status != NC_NOERR) ERR(status);

            status = nc_put_vara_double(grp_sds, varid[14], start, count, lonln);
            if(status != NC_NOERR) ERR(status);
        
            status = nc_put_vara_double(grp_sds, varid[15], start, count, east);
            if(status != NC_NOERR) ERR(status);

            status = nc_put_vara_double(grp_sds, varid[16], start, count, north);
            if(status != NC_NOERR) ERR(status);

            status = nc_put_vara_double(grp_sds, varid[17], start, count, elev);
            if(status != NC_NOERR) ERR(status);

            // Save lat/lon for L2
            offset = i * Samples;
            memcpy( lat+offset, latln, lineSize);
            memcpy( lon+offset, lonln, lineSize);
            //printf("lat+offset = %f, lon+offset = %f\n", *(lat+offset), *(lon+offset) );

            /* Alternative to memcpy
            for(j = 0; j < Samples; j++){
                *(lat + i*Samples + j) = latln[j];
                *(lon + i*Samples + j) = lonln[j];
                if(i == 10 && j == 0){
                    //printf("*(lat + 10*Samples) = %f\n", *(lat + i*Samples + j));
                    //printf("*(lon + 10*Samples) = %f\n", *(lon + i*Samples + j));
                 }
            }*/
        }
        else if(level == 2){
            status = nc_put_vara_double(grp_sds, varid[2], start, count, latln);
            if(status != NC_NOERR) ERR(status);

            status = nc_put_vara_double(grp_sds, varid[3], start, count, lonln);
            if(status != NC_NOERR) ERR(status);
        }
    } //end for(i = ...)

    free(east); free(north); free(latln); free(lonln); free(elev); free(tempBuf);
    return 1;
}

//===========================================
void writeLatLon(int grp_sds, int lines, int *varid, double *lat, double *lon)
{ // Used in L2 only
    int i, status;
    double *latln = calloc(Samples, sizeof(double));
    double *lonln = calloc(Samples, sizeof(double));
    size_t lineSize = Samples*sizeof(double);
    size_t offset;

    size_t start[2], count[2];
    count[0] = 1;
    count[1] = Samples;
    start[1] = 0;

    for(i = 0; i < lines; i ++) {
        start[0] = i;
        offset = i*lineSize;

        memcpy(latln, lat+Samples*i, lineSize);
        memcpy(lonln, lon+Samples*i, lineSize);

        status = nc_put_vara_double(grp_sds, varid[2], start, count, latln);
        if(status != NC_NOERR) ERR(status);

        status = nc_put_vara_double(grp_sds, varid[3], start, count, lonln);
        if(status != NC_NOERR) ERR(status);

    }

    free(latln); free(lonln);
}

//===========================================
void tm_to_latlon(int zone, int hemi, double* lat, double* lon, double* x, double* y, int pix)
{
//
// convert UTM to lat/lon
//
// a = ellipsoid major axis (meters)
// b = ellipsoid minor axis (meters)
// lat0 = origin latitude (degrees)
// lon0 = origin longitude (degrees) (W is negative)
// x0 = false easting (meters)
// y0 = false northing (meters)
// lat,lon = geographic coords (degrees) (W is negative)
// x,y = map coordinates (meters)
//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
//  if (n_elements(zone) eq 0) then $
//    zone = fix(31. + lon[0]/6.)
//  ; utm default values:
//  lat0 = 0d
//  lon0 = 3d + (zone - 31) * 6d
//  a = 6378206.4d ; clarke 1866 ellipsoid (a)
//  b = 6356583.8d ; clarke 1866 ellipsoid (b)
//  k0 = .9996d
//  x0 = 500000d
//  y0 = 0d or 10000000d for north/south
//
  int  i;
  double wgs_a = 6378137.0; //ellipsoid major axis (meters)
  double wgs_b = 6356752.3; //ellipsoid minor axis (meters)
  double lat0 = 0.0;
  double lon0 =3.0 + (zone - 31)*6.0;
  double k0 = .9996;
  double x0 = 500000.0;     //false easting (meters)
  double y0 = 0.0;
  double e2, e4, e6, lat0r, lon0r, s2lat0r, s4lat0r, s6lat0r;
  double m0, ep2, e1, e12, e13, e14;
  double delx, dely, d1, latr, lonr, m, mu, lat1, clat1, slat1, tlat1;
  double c1, t1, n1, r1, konst1;

  if( hemi == 1 ) y0 = 0.0;
  else            y0 = 10000000.0;

  e2 = (wgs_a*wgs_a-wgs_b*wgs_b) / (wgs_a*wgs_a);
  e4 = e2*e2;
  e6 = e4*e2;
  lat0r = DPI*lat0/180.0;
  lon0r = DPI*lon0/180.0;
  s2lat0r = sin(2.0*lat0r);
  s4lat0r = sin(4.0*lat0r);
  s6lat0r = sin(6.0*lat0r);
  m0 = (1.-e2/4.-3.*e4/64.-5.*e6/256.)*lat0r;
  m0 = m0-(3.*e2/8.+3.*e4/32.+45.*e6/1024.)*s2lat0r;
  m0 = m0+(15.*e4/256.+45.*e6/1024)*s4lat0r;
  m0 = m0-(35.*e6/3072)*s6lat0r;
  m0 = wgs_a*m0;
  ep2 = e2/(1.0-e2);
  e1 = (1.0-sqrt(1.0-e2))/(1.0+sqrt(1.0-e2));
  e12 = e1*e1;
  e13 = e12*e1;
  e14 = e13*e1;
  konst1 = 180.0/DPI;

  for( i = 0; i < pix; i++ ) {
    delx = x[i]-x0;
    dely = y[i]-y0;
    m    = m0+dely/k0;
    mu   = m/(wgs_a*(1.0-e2/4.0-3.0*e4/64.0-5.0*e6/256.0));
    lat1 = mu+(3.0*e1/2.0-27.0*e13/32.0)*sin(2.0*mu)+
         (21.0*e12/16.0-55.0*e14/32.0)*sin(4.0*mu)+(151.0*e13/96.0)*sin(6.0*mu);
    clat1 = cos(lat1);
    slat1 = sin(lat1);
    tlat1 = tan(lat1);
    c1    = ep2*clat1*clat1;
    t1    = tlat1*tlat1;
    n1    = wgs_a/sqrt(1.0-e2*slat1*slat1);
    r1    = wgs_a*(1.0-e2)/( pow( ( 1.0-e2*slat1*slat1), 1.5 ) );
    d1    = delx/(n1*k0);
    latr  = lat1 - (n1*tlat1/r1)*(d1*d1/2.0 -
            (5.0+3.0*t1+10.0*c1-4.0*c1*c1-9.0*ep2)*pow( d1, 4.0 )/24.0 +
        (61+90*t1+298*c1+45*t1*t1-252*ep2-3*c1*c1)*pow( d1, 6.0 )/720.0);

    lonr  = lon0r + (d1-(1.0+2.0*t1+c1)*pow( d1, 3.0 )/6.0 +
      (5.0-2.0*c1+28.0*t1-3.0*c1*c1+8.0*ep2+24.0*t1*t1)*pow( d1, 5.0 )/120.0)/clat1;

    lat[i] = latr*konst1;
    lon[i] = lonr*konst1;
//if(i==0) printf("x[0] = %f, y[0] = %f, lat[0] = %f, lon[0] = %f\n", x[0], y[0], lat[0], lon[0]);
  }
}


//===========================================
off_t getGranuleSize(char *filename)
{
    int status;
    struct stat stats;
    FILE *stream;  // binary input file pointer (return value)
    off_t filesize = 0;

    // File size is greater than 2G...cannot use fseek/ftell
    if(stat(filename, &stats) == 0){
        filesize = stats.st_size;
    }
    else printf("Unable to get file size\n");

    return filesize;
}

//===========================================
int minLines(FILE *rxxhdr, FILE *igmhdr, FILE* obshdr)
{
    int rxx_lines=0, igm_lines=0, obs_lines=0;
    int lines, diff=0;

    if(rxxhdr != NULL) rxx_lines = getLines(rxxhdr);

    if(igmhdr != NULL) igm_lines = getLines(igmhdr);

    if(obshdr != NULL) obs_lines = getLines(obshdr);

    if(rxx_lines > 0 && igm_lines > 0 && rxx_lines != igm_lines ) {
        printf("*** WARNING: IGM(%d) and RND/RFL(%d) files have differnt number of lines! ***\n", igm_lines, rxx_lines);
        diff  = 1;
    }
    if(rxx_lines > 0 && obs_lines > 0 && rxx_lines != obs_lines) {
        printf("*** WARNING: OBS(%d) and RDN/RFL(%d) files have differnt number of lines! ***\n", obs_lines, rxx_lines);
        diff = 1;
    }
    if(igm_lines > 0 && obs_lines > 0 && igm_lines != obs_lines) {
        printf("*** WARNING: IGM(%d) and OBS(%d) files have differnt number of lines! ***\n", igm_lines, obs_lines);
        diff = 1;
    }

    if(diff){
        lines = MIN(rxx_lines, igm_lines);
        lines = MIN(lines, obs_lines);
        printf("*** WARNING: The output will be %d lines long ***\n", lines);
    }
    else lines = rxx_lines;

    return lines;
}
